from countergenedit.editing.activation_utils import get_mlp_modules, get_res_modules
from countergenedit.editing.activation_ds import ActivationsDataset
from countergenedit.editing.edition import edit_model, ReplacementConfig, get_edit_configs



import copy
from typing import Dict, Iterable, List
from countergenedit.tools.math_utils import project
from transformers import GPT2Model
from torch import nn
import torch
from attrs import define


def get_edit_configs(named_modules: Dict[str, nn.Module], dirs: torch.Tensor, has_leftover: bool = False):
    """Return the configs where the same projections along dirs is done at the output of each module.

    The keys of the named_modules should be the name of the corresponding module in the original model."""
    return [ReplacementConfig(name, module, dirs, has_leftover) for name, module in named_modules.items()]


@define
class ReplacementConfig:
    """Configuration for an edition by projection."""

    module_name: str  #: The name of the module in the original network you wish to replace
    old_module: nn.Module  #: The module object to replace
    dirs: torch.Tensor  #: A 2D Float Tensor of shape (n, hidden_dim) listing vectors along which to project
    #: If True, the output of the module is expected to be (to_proj, sth...) rather than to_proj.
    has_leftover: bool = False


def edit_model(model: nn.Module, configs: Iterable[ReplacementConfig]) -> nn.Module:
    """Return a new module where the replacements described in the config have been done."""
    model = copy.deepcopy(model)
    for config in configs:
        new_module = ProjectionWrapper(config.old_module, config.dirs, config.has_leftover)

        *parent_path, name = config.module_name.split(".")
        parent_name = ".".join(parent_path)
        parent = model.get_submodule(parent_name)
        if hasattr(parent, name):  # Regular case, if it's a regular attribute
            setattr(parent, name, new_module)
        else:  # ModuleList case, if it's the member of a list
            parent[int(name)] = new_module  # type: ignore
    return model


class ProjectionWrapper(nn.Module):
    def __init__(self, wrapped_module: nn.Module, dirs: torch.Tensor, has_leftover: bool = False):
        super().__init__()
        self.wrapped_module = wrapped_module.to(dirs.device)
        self.dirs = dirs
        self.has_leftover = has_leftover

        if not torch.allclose(
            torch.eye(self.dirs.shape[0]).to(dirs.device), torch.einsum("m k, n k -> m n", dirs, dirs), atol=1e-4
        ):
            raise ValueError("Directions should be orthonrmal")

    def forward(self, *args, **kwargs):
        y = self.wrapped_module(*args, **kwargs)

        if self.has_leftover:
            hidden_states, *leftover = y
        else:
            hidden_states = y

        # hidden_states -= torch.einsum("b n h, m h, m k -> b n k", hidden_states, self.dirs, self.dirs)
        hidden_states = project(hidden_states, self.dirs)

        return (hidden_states, *leftover) if self.has_leftover else hidden_states



from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Sequence, TypeVar, Tuple
import torch


def project(dir: torch.Tensor, dirs: torch.Tensor) -> torch.Tensor:
    """Return dir, but projected in the orthogonal of the subspace spanned by dirs.

    Assume that dirs are already orthonomal, and that the number of dimensions is > 0."""
    inner_products = torch.einsum("n h, ...h -> ...n", dirs, dir)
    new_dir = dir - torch.einsum("...n, n h -> ...h", inner_products, dirs)

    return new_dir

def orthonormalize(dirs: torch.Tensor) -> torch.Tensor:
    """Apply the Gram-Schmidt algorithm to make dirs orthonormal
    
    Assumes that the number of dimensions and dirs is > 0."""
    n, _ = dirs.shape

    dirs[0] /= torch.linalg.norm(dirs[0])
    for i in range(1, n):
        dirs[i] = project(dirs[i], dirs[:i])
        dirs[i] /= torch.linalg.norm(dirs[i])
    
    return dirs


from math import exp
from typing import TYPE_CHECKING, Callable, Sequence, List, Optional, Tuple

import torch
from countergen.tools.math_utils import perplexity
from countergenedit.tools.utils import concat_dicts, get_device, get_gpt_tokenizer, remove_last_tok
from countergen.types import Input, ModelEvaluator, Outputs, Performance
from countergen import get_generative_model_evaluator
from transformers import BatchEncoding, GPT2LMHeadModel, GPT2Tokenizer

metrics = ["perplexity", "probability"]

LogProbs = float

# Return a log prob for each token of output
GenerativeModel = Callable[[Input, Outputs], Sequence[Sequence[LogProbs]]]


def pt_to_generative_model(model: torch.nn.Module, device: Optional[str] = None) -> GenerativeModel:
    """Make a GenerativeModel out of a pytorch model.

    The model should take {"input_ids": [tensor], "attention_mask": [tensor]} as input,
    and return something that has a "logits" attribute."""

    tokenizer = get_gpt_tokenizer()
    device = device or get_device()
    model = model.to(device)

    def gen_model(inp: Input, out: Outputs) -> List[List[float]]:
        tokens_inp = tokenizer(inp, return_tensors="pt").to(device)
        token_outs = [tokenizer(o, return_tensors="pt").to(device) for o in out]

        return get_correct_logprobs(tokens_inp, token_outs, model)

    return gen_model


def get_correct_logprobs(
    tokens_inp: BatchEncoding, token_outs: List[BatchEncoding], model: torch.nn.Module
) -> List[List[float]]:

    if all([o["input_ids"].shape[-1] == 1 for o in token_outs]):
        return [[x.item() for x in t] for t in get_correct_1tok_logprobs(tokens_inp, token_outs, model)]

    inp_length = tokens_inp["input_ids"].shape[-1]

    result: List[List[float]] = []

    for tokens_out in token_outs:
        out_length = tokens_out["input_ids"].shape[-1]
        assert out_length > 0, "Zero length expected output is forbidden"

        tokens_to_feed = remove_last_tok(concat_dicts([tokens_inp, tokens_out]))
        with torch.no_grad():
            logits = model(**tokens_to_feed).logits[0]
        log_probs = torch.log_softmax(logits, dim=-1)[inp_length - 1 :, :]

        assert len(log_probs) == len(tokens_out["input_ids"][0])
        correct_log_probs = torch.gather(log_probs.to("cpu"), 1, tokens_out["input_ids"][0, :, None].to("cpu"))[:, 0]

        result.append([x.item() for x in correct_log_probs])

    return result


def get_correct_1tok_logprobs(
    tokens_inp: BatchEncoding, token_outs: List[BatchEncoding], model: torch.nn.Module
) -> List[torch.Tensor]:

    with torch.no_grad():
        logits = model(**tokens_inp).logits[0].to("cpu")
    log_probs = torch.log_softmax(logits, dim=-1)[-1:, :]

    good_tokens = [o["input_ids"][0, 0].item() for o in token_outs]

    correct_log_probs = [log_probs[:, i] for i in good_tokens]

    return correct_log_probs



from functools import lru_cache
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Sequence, Tuple, TypeVar

import torch
from transformers import GPT2Tokenizer

T = TypeVar("T")


def concat_dicts(dicts: Sequence[Mapping[Any, torch.Tensor]]) -> Dict[Any, torch.Tensor]:
    if not dicts:
        raise ValueError("dicts is empty")
    keys = dicts[0].keys()
    for d in dicts:
        if d.keys() != keys:
            raise ValueError("dicts must have the same keys")
    return {k: torch.cat([d[k] for d in dicts], dim=-1) for k in keys}


def remove_last_tok(d: Mapping[Any, torch.Tensor]) -> Dict[Any, torch.Tensor]:
    return {k: t[:, :-1] for k, t in d.items()}


def get_device() -> str:
    return "cuda:0" if torch.cuda.is_available() else "cpu"


@lru_cache(maxsize=1)
def get_gpt_tokenizer() -> GPT2Tokenizer:
    return GPT2Tokenizer.from_pretrained("gpt2")
